# from translator import i18n
from . import translator
trans = translator.i18n

bl_info = {
    "name" : "w-节点组输入助手(Group input helper)-添加拆分合并移动(Add Split Merge Move)",
    "author" : "一尘不染",
    "description" : "快速添加组输入节点-拆分合并移动组输入节点-快速添加组输入输出接口(Qucik add and split merge move Group Input node-Qucik add Group Input Output socket)",
    "blender" : (3, 0, 0),
    "version" : (2, 3, 0),
    "location" : "偏好设置-标题栏-N面板(Preferences-Editor Bar-N panel)",
    "warning" : "",
    "doc_url": "",
    "tracker_url": "",
    "category" : "Node"
}

import bpy
from bpy.types import Operator, Menu, Panel, AddonPreferences

import bpy.utils.previews
from mathutils import Vector
import os

addon_keymaps = {}
_icons = None

socket_name = [
            "Geometry",
            "Shader",
            "Boolean",
            "Float",
            "Integer",
            "Vector",
            "Color",
            "Rotation",
            "Matrix",
            "Menu",
            "String",
            "Material",
            "Object",
            "Collection",
            "Image",
            "Texture",
            ]
socket_name_cn = [
            "几何数据",
            "着色器",
            "布尔",
            "浮点",
            "整数",
            "矢量",
            "颜色",
            "旋转",
            "矩阵",
            "菜单",
            "字符串",
            "材质",
            "物体",
            "集合",
            "图像",
            "纹理",
            ]
socket_bl_socket_idname = [
            "NodeSocketGeometry",
            "NodeSocketShader",
            "NodeSocketBool",
            "NodeSocketFloat",
            "NodeSocketInt",
            "NodeSocketVector",
            "NodeSocketColor",
            "NodeSocketRotation",
            "NodeSocketMatrix",
            "NodeSocketMenu",
            "NodeSocketString",
            "NodeSocketMaterial",
            "NodeSocketObject",
            "NodeSocketCollection",
            "NodeSocketImage",
            "NodeSocketTexture",
            ]
png_list = [
            '几何数据.png',
            '着色器.png',    # Shader接口也用几何数据
            '布尔.png',
            '浮点.png',
            '整数.png',
            '矢量.png',
            '颜色.png',
            '旋转.png',
            '矩阵.png',
            '菜单.png',
            '字符串.png',
            '材质.png',
            '物体.png',
            '集合.png',
            '图像.png',
            '纹理.png',
            '空.png',
            ]
Geometry_socket_type = [   # 这里没用到
            "GEOMETRY",
            "SHADER",
            "BOOLEAN",
            "VALUE",
            "INT",
            "VECTOR",
            "RGBA",
            "ROTATION",
            "MATRIX",
            "MENU",
            "STRING",
            "MATERIAL",
            "OBJECT",
            "COLLECTION",
            "IMAGE",
            "TEXTURE",
            "CUSTOM",    # 黑色虚拟接口
            ]

inputs_png = {k:v for k, v in zip(socket_bl_socket_idname, png_list)}
# inputs_png["NodeSocketShader"] = '几何数据.png'
get_socket_name = {k:v for k, v in zip(socket_bl_socket_idname, socket_name_cn)}
# get_socket_name = {k:v for k, v in zip(socket_bl_socket_idname, zip(socket_name_cn, socket_name))}

def find_user_keyconfig(key):
    km, kmi = addon_keymaps[key]
    for item in bpy.context.window_manager.keyconfigs.user.keymaps[km.name].keymap_items:
        found_item = False
        if kmi.idname == item.idname:
            found_item = True
            for name in dir(kmi.properties):
                if not name in ["bl_rna", "rna_type"] and not name[0] == "_":
                    if not kmi.properties[name] == item.properties[name]:
                        found_item = False
        if found_item:
            return item
    print(f"Couldn't find keymap item for {key}, using addon keymap instead. This won't be saved across sessions!")
    return kmi

# INVOKE_REGION_WIN                                                2 新建节点 唤出菜单新建跟随鼠标，标题栏新建跟随
# INVOKE_DEFAULT  INVOKE_REGION_CHANNELS  INVOKE_REGION_PREVIEW    1 新建节点 唤出菜单新建跟随鼠标，标题栏新建不跟随
# INVOKE_AREA  INVOKE_SCREEN                                       报错
# EXEC_DEFAULT  EXEC_REGION_WIN  EXEC_AREA  EXEC_SCREEN  EXEC_REGION_CHANNELS  EXEC_REGION_PREVIEW  0 默认在鼠标位置新建

class W_AddonPreferences(AddonPreferences):
    bl_idname = __name__
    show_panel_name: bpy.props.BoolProperty(name='显示面板名字',  description=trans('添加组输入菜单里显示接口所属面板名字'), default=False)
    simplify_menu:   bpy.props.BoolProperty(name='simplify_menu', description=trans('简化<组输入合并拆分移动>菜单'), default=True)
    def draw(self, context):
        layout = self.layout
        layout.label(text=trans('标题栏和N面板显示组输入拆分'), icon="RADIOBUT_ON")

        split1 = layout.split(factor=0.4, align=True)
        split1.label(text=trans('添加组输入-菜单'))
        split1.prop(find_user_keyconfig('A6EHS'), 'type', text='', full_event=True)

        split2 = layout.split(factor=0.4, align=False)
        split2.label(text=trans('添加组输入-面板'))
        split2.prop(find_user_keyconfig('key_PT_Add_Group_Input'), 'type', text='', full_event=True)

        split3 = layout.split(factor=0.4, align=True)
        split3.label(text=trans('添加组输入输出接口'))
        split3.prop(find_user_keyconfig('key_Add_New_Group_Item'), 'type', text='', full_event=True)

        split3 = layout.split(factor=0.4, align=True)
        split3.label(text=trans('组输入合并拆分移动'))
        split3.prop(find_user_keyconfig('key_Merge_Split_Move_Group_Input'), 'type', text='', full_event=True)

        split4 = layout.split(factor=0.4)
        split4.label(text=trans('显示面板名字'))
        split4.prop(self, 'show_panel_name', text='')
        
        split5 = layout.split(factor=0.4)
        split5.label(text=trans('简化<组输入合并拆分移动>菜单'))
        split5.prop(self, 'simplify_menu', text='')

def w_add_to_node_mt_editor_menus_45569(self, context):
    layout = self.layout
    layout.menu('W_MT_Add_Group_Input_Hided_Socket', text=trans('组输入'))

def get_socket_icon(layout):
    tree = bpy.context.space_data.edit_tree
    # if bpy.data.version < (4, 0, 0):    # 为啥alpha版version和3.6一样都是(3, 6, 11) 找了好久 app.version正确
    # if bpy.app.version < (4, 0, 0)
    if hasattr(tree, "inputs"):
        try:
            group_inputs = tree.inputs
        except:
            pass
    if hasattr(tree, "interface"):
        try:
            group_inputs = []
            for item in tree.interface.items_tree:
                if item.item_type == 'SOCKET' and item.in_out == 'INPUT':
                    group_inputs.append(item)
        except:
            pass
    for index, group_input in enumerate(group_inputs):
        if hasattr(group_input, "position"):
            if group_input.position == 0:
                layout.separator()
                prefs = bpy.context.preferences.addons[__name__].preferences
                if prefs.show_panel_name:
                    layout.label(text=group_input.parent.name, icon="MENU_PANEL")
                    layout.separator()
        socket_id = group_input.bl_socket_idname
        # 好像只3.6这样,4.0就都一种了
        if socket_id.startswith("NodeSocketFloat"):
            socket_id = "NodeSocketFloat"
        if socket_id.startswith("NodeSocketInt"):
            socket_id = "NodeSocketInt"
        if socket_id.startswith("NodeSocketVector"):
            socket_id = "NodeSocketVector"
        input_png = inputs_png[socket_id]
        socket_name = group_input.name
        if socket_name == "":
            socket_name = " "
        op = layout.operator('w.add_group_input_hided_socket', text=socket_name, icon_value=(_icons[input_png].icon_id ) )
        # op.bl_description = group_input.description    # 不知道为啥不对
        op.menu_index = index
    layout.separator()
    op = layout.operator('w.add_group_input_hided_socket', text=trans('空'), icon_value=_icons['空.png'].icon_id)
    op.menu_index = -1

def get_icon_add_new_socket(layout, context):
    for socket_type, socket_name in get_socket_name.items():
        input_png = inputs_png[socket_type]
        tree_type = context.space_data.edit_tree.bl_idname
        group_type = None
        a_node = context.active_node
        name = trans(socket_name)
        if a_node and a_node.select:
            group_type = a_node.bl_idname
        # Todo 拆分后删除转接口
        # Todo 顶层材质不显示着色器
        # Todo 着色器接口排在最上面
        # _ 选定节点组添加节口时额外提示
        # Todo 看心情添加版本控制
        # Todo 百度网盘更新
        # print("--" * 20)
        # print(f"{tree_type = }")
        # print(f"{a_node.bl_idname = }")
        # print(f"{a_node.type = }")
        # if tree_type == "ShaderNodeTree" or group_type == "GROUP":            # 这样对于组节点就绘制多次了
        if tree_type == "ShaderNodeTree" or group_type == "ShaderNodeGroup":
            if socket_name in ["着色器", "布尔", "浮点", "整数", "矢量", "颜色"]:
                op = layout.operator('w.add_new_group_item', text=name, icon_value=(_icons[input_png].icon_id ) )
                op.socket_type = socket_type
        if tree_type == "CompositorNodeTree" or group_type == "CompositorNodeGroup":
            if socket_name in ["浮点", "矢量", "颜色"]:
                op = layout.operator('w.add_new_group_item', text=name, icon_value=(_icons[input_png].icon_id ) )
                op.socket_type = socket_type
        if tree_type == "GeometryNodeTree" and socket_name != "着色器":
        # if group_type == "GeometryNodeGroup" or tree_type == "GeometryNodeTree":
            op = layout.operator('w.add_new_group_item', text=name, icon_value=(_icons[input_png].icon_id ) )
            op.socket_type = socket_type

def abs_loc(node):
    return node.location + abs_loc(node.parent) if node.parent else node.location

def get_nodes_center(nodes):
    if nodes:     #  没有节点不运行
        locs = []
        for node in nodes:
            node.select = True
            if node.bl_idname == "NodeFrame": continue
            location = abs_loc(node)
            # center = location + Vector((node.width / 2, -node.dimensions.y / 2))
            center = [location.x, location.x + node.width, location.y, location.y -node.dimensions.y]
            locs.append(center)
        print([[l[0], l[1]] for l in locs])
        # # 选中节点重心位置
        # center_x = sum(x for x, y in locs) / len(nodes); center_y = sum(y for x, y in locs) / len(nodes)
        # # 选中节点中心位置
        # center_x1 = (max(x for x, y in locs) + min(x for x, y in locs)) / 2: center_y1 = (max(y for x, y in locs) + min(y for x, y in locs)) / 2
        center_x1 = (min(l[0] for l in locs) + max(l[1] for l in locs)) / 2
        center_y1 = (max(l[2] for l in locs) + min(l[3] for l in locs)) / 2
        nodes_center = Vector((center_x1, center_y1))
        return nodes_center

def has_link(node):
    for socket in node.outputs:
        if socket.is_linked:
            return True
    return False

def is_tall(socket):
    if socket.type not in ["VECTOR", "ROTATION"]:
        return False
    if socket.hide_value:
        return False
    if socket.is_linked:
        return False
    return True

def get_in_socket_location(tar_node, tar_socket):
    has_tall = 0
    for socket in reversed(tar_node.inputs):
        has_tall += is_tall(socket)
        print(f"{has_tall=}")
    deviation = 5 * has_tall if has_tall else 0      # 矢量接口导致dimensions.y不准?
    x = abs_loc(tar_node).x - 190
    y = abs_loc(tar_node).y - tar_node.dimensions.y + deviation + 43     # 额外加的是组输入输出接口和顶端的距离差
    soc_loc_dict = {}    # socket_location_dictionary
    for socket in reversed(tar_node.inputs):
        if socket.hide or not socket.enabled:
            continue
        y_offset = 21.9   # 接口之间常规距离
        vec_offset = 60   # 矢量接口距离
        y += y_offset + vec_offset * is_tall(socket)
        soc_loc_dict[socket] = y
    return Vector((x, soc_loc_dict[tar_socket]))

class W_OT_Add_Group_Input_Hided_Socket(Operator):
    bl_idname = "w.add_group_input_hided_socket"
    bl_label = trans("组输入隐藏节口")
    bl_description = trans("组输入拆分菜单: 快捷键 1 和 Ctrl Shift Alt 1")
    bl_options = {"REGISTER", "UNDO"}
    menu_index: bpy.props.IntProperty(name='menu index', description='', default=0, subtype='NONE')

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        bpy.ops.node.add_node('INVOKE_REGION_WIN', use_transform=True, type='NodeGroupInput')
        # bpy.ops.node.add_node('INVOKE_REGION_WIN', type='NodeGroupInput')
        a_node = bpy.context.active_node
        # a_node.location = context.space_data.cursor_location
        # a_node.location += Vector((-140, 35))
        # bpy.ops.transform.trans('INVOKE_DEFAULT')

        Variable = self.menu_index
        index = -1
        for output in a_node.outputs:
            if Variable == -1:
                output.hide = True
                if output.name == "" and output.type == "CUSTOM":
                    output.hide = False
            else:
                index += 1   # +1 要放到上面不能在下面，因为每次循环index+1 在下面一旦满足条件之后就一直continue
                if index == Variable:
                    continue
                output.hide = True

        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Add_New_Group_Item(Operator):
    bl_idname = "w.add_new_group_item"
    bl_label = trans("添加输入输出接口")
    bl_description = trans("添加输入输出接口")
    bl_options = {"REGISTER", "UNDO"}
    socket_type: bpy.props.StringProperty(name='socket type', description='')

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        a_node = context.active_node
        if a_node and a_node.select and a_node.bl_idname in ["GeometryNodeGroup", "ShaderNodeGroup"]:
            tree = a_node.node_tree
        else:
            tree = context.space_data.edit_tree
        interface = tree.interface
        name = trans(get_socket_name[self.socket_type])
        if context.scene.add_input_socket:
            item = interface.new_socket(name, socket_type=self.socket_type, in_out="INPUT")
            interface.active = item
        if context.scene.add_output_socket:
            item = interface.new_socket(name, socket_type=self.socket_type, in_out="OUTPUT")
            interface.active = item
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Merge_Group_Input_Socket(Operator):
    bl_idname = "cloud.merge_group_input_socket"
    bl_label = trans("合并组输入接口")
    bl_description = trans("合并组输入接口,留下未隐藏的接口")
    bl_options = {"REGISTER", "UNDO"}
    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        tree = context.space_data.edit_tree
        nodes = tree.nodes
        a_node = nodes.active
        # nodes = context.selected_nodes
        # a_node = context.active_node
        links = tree.links
        list_group = []
        for node in nodes:
            if node.bl_idname == "NodeGroupInput" and node.select:
                list_group.append(node)
        if len(list_group) >= 2:
            condition = a_node and a_node.bl_idname == "NodeGroupInput" and a_node.select
            if condition:
                target_group = a_node
            else:
                target_group = list_group[0]
                nodes_center = get_nodes_center(list_group)
            group_id = {socket.identifier: socket for socket in target_group.outputs}
            for node in list_group:
                if node != target_group:
                    for out_socket in node.outputs:
                        tar_socket = group_id[out_socket.identifier]
                        if out_socket.is_linked:
                            for link in out_socket.links:
                                links.new(tar_socket, link.to_socket)
                        if out_socket.hide == False:
                            tar_socket.hide = False
                    nodes.remove(node)    # 没出错,不确定这样删会不会出错(删的是nodes里的)
            if not condition:
                tar_loc = target_group.dimensions
                target_group.location = nodes_center - Vector((tar_loc.x / 2,  - tar_loc.y))
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

def merge_group_input_linked(selected_nodes, active_node, loc_at_max_y=False):
    tree = bpy.context.space_data.edit_tree
    nodes = tree.nodes
    links = tree.links
    list_group = []
    list_group_loc_y = []
    for node in selected_nodes:
        if node.bl_idname == "NodeGroupInput":
            list_group.append(node)
            list_group_loc_y.append(abs_loc(node).y)
    list_group_loc_y.sort()
    y_max = list_group_loc_y[-1]
    if len(list_group) == 1: 
        return None
    condition = active_node and active_node.bl_idname == "NodeGroupInput" and active_node.select
    if condition:
        target_group = active_node
    else:
        target_group = list_group[0]
        nodes_center = get_nodes_center(list_group)
    group_id = {out_socket.identifier: out_socket for out_socket in target_group.outputs}
    for node in list_group:
        if node != target_group:
            for out_socket in node.outputs:
                if out_socket.is_linked:
                    for link in out_socket.links:
                        links.new(group_id[out_socket.identifier], link.to_socket)
            nodes.remove(node)    # 没出错,不确定这样删会不会出错
    if not condition:
        tar_loc = target_group.dimensions
        target_group.location = nodes_center - Vector((tar_loc.x / 2, - tar_loc.y))
    if loc_at_max_y:
        target_group.location.y = y_max
    for soc in target_group.outputs:
        soc.hide = True
    return target_group

class W_OT_Merge_Group_Input_Linked(Operator):
    bl_idname = "cloud.merge_group_input_linked"
    bl_label = trans("合并组输入连线")
    bl_description = trans("合并组输入连线,只留下连线的接口")
    bl_options = {"REGISTER", "UNDO"}
    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        selected_nodes = context.selected_nodes
        active_node = context.active_node
        merge_group_input_linked(selected_nodes, active_node)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Hide_Group_Input_Sockets(Operator):
    bl_idname = "cloud.hide_group_input_sockets"
    bl_label = trans("隐藏组输入接口")
    bl_description = trans("隐藏组输入节点未使用的接口")
    bl_options = {"REGISTER", "UNDO"}
    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        nodes = context.space_data.edit_tree.nodes
        for node in nodes:
            if node.bl_idname == "NodeGroupInput":
                for soc in node.outputs:
                    soc.hide = True

        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Split_Group_Input_Socket(Operator):
    bl_idname = "cloud.split_group_input_socket"
    bl_label = trans("拆分组输入接口")
    bl_description = trans("拆分组输入接口,留下未隐藏的接口")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        tree = context.space_data.edit_tree
        nodes = tree.nodes
        links = tree.links
        selected_nodes = context.selected_nodes
        i = 0
        for node in selected_nodes:
            if node.bl_idname == "NodeGroupInput" and node.select:
                i = -1
                for out_soc in node.outputs:
                    if not out_soc.hide and out_soc.enabled and out_soc.bl_idname != "NodeSocketVirtual":
                    # if out_soc.is_linked:
                        i += 1
                        group_node = nodes.new('NodeGroupInput')
                        group_node.location = abs_loc(node)
                        group_node.location.y = abs_loc(node).y - 80 * i
                        group_id = {socket.identifier: socket for socket in group_node.outputs}
                        for link in out_soc.links:
                            links.new(group_id[out_soc.identifier], link.to_socket)
                        for new_out_soc in group_node.outputs:
                            if new_out_soc.identifier != out_soc.identifier:
                                new_out_soc.hide = True
                nodes.remove(node)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Split_All_Group_Input_Socket(Operator):
    bl_idname = "cloud.split_all_group_input_socket"
    bl_label = trans("拆分组输入接口")
    bl_description = trans("拆分组输入接口,留下未隐藏的接口,一连多也拆开")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        tree = context.space_data.edit_tree
        nodes = tree.nodes
        links = tree.links
        selected_nodes = context.selected_nodes
        i = 0
        for node in selected_nodes:
            if node.bl_idname == "NodeGroupInput" and node.select:
                i = -1
                for out_soc in node.outputs:
                    if not out_soc.hide and out_soc.enabled and out_soc.bl_idname != "NodeSocketVirtual":
                        link_count = len(out_soc.links)
                        soc_links = out_soc.links if out_soc.links else range(1)
                        for link in soc_links:
                            if hasattr(link, "is_valid") and not link.is_valid:       # 有些线存在，但是因为节点不同选项，线不可见
                                continue
                            i += 1
                            group_node = nodes.new('NodeGroupInput')
                            group_node.location = abs_loc(node)
                            group_node.location.y = abs_loc(node).y - 80 * i
                            group_id = {socket.identifier: socket for socket in group_node.outputs}
                            if link_count:
                                links.new(group_id[out_soc.identifier], link.to_socket)
                            for new_out_soc in group_node.outputs:
                                if new_out_soc.identifier != out_soc.identifier:
                                    new_out_soc.hide = True
                nodes.remove(node)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Split_All_And_Move(Operator):
    bl_idname = "cloud.split_all_and_move"
    bl_label = trans("拆分组输入并移动")
    bl_description = trans("拆分组输入接口,并移动到连向接口的附近,留下未隐藏的接口,一连多也拆开")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        tree = context.space_data.edit_tree
        nodes = tree.nodes
        links = tree.links
        selected_nodes = context.selected_nodes
        i = 0
        for node in selected_nodes:
            if node.bl_idname == "NodeGroupInput" and node.select and has_link(node):
                i = -1
                for out_soc in node.outputs:
                    if not out_soc.hide and out_soc.enabled and out_soc.bl_idname != "NodeSocketVirtual":
                        # 删掉组输入输出接口后面连的所有转接点
                        if out_soc.links:
                            for link in out_soc.links:
                                delete_reroute(link, nodes, links, out_soc)
                            
                    if not out_soc.hide and out_soc.enabled and out_soc.bl_idname != "NodeSocketVirtual":
                        link_count = len(out_soc.links)
                        soc_links = out_soc.links if out_soc.links else range(1)
                        for link in soc_links:
                            if hasattr(link, "is_valid") and not link.is_valid:       # 有些线存在，但是因为节点不同选项，线不可见
                                continue
                            i += 1
                            group_node = nodes.new('NodeGroupInput')
                            # group_node.location = node.location
                            # group_node.location.y = node.location.y - 80 * i
                            group_id = {socket.identifier: socket for socket in group_node.outputs}
                            spec_in_socket = group_id[out_soc.identifier]
                            if link_count:         # is_linked:
                                links.new(spec_in_socket, link.to_socket)
                                to_node = spec_in_socket.links[0].to_node
                                to_socket = spec_in_socket.links[0].to_socket
                                group_node.location = get_in_socket_location(to_node, to_socket)
                            for new_out_soc in group_node.outputs:
                                if new_out_soc.identifier != out_soc.identifier:
                                    new_out_soc.hide = True
                nodes.remove(node)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

def delete_reroute(link, nodes, links, out_socket):
    if link.to_node.type == "REROUTE":
        reroute = link.to_node
        for re_link in reroute.outputs[0].links:
            links.new(out_socket, re_link.to_socket)
            delete_reroute(re_link, nodes, links, out_socket)
        nodes.remove(reroute)
# def delete_reroute(link, nodes, links, out_socket):
#     if link.to_node.type == "REROUTE":
#         reroute = link.to_node
#         re_links = reroute.outputs[0].links
#         for re_link in re_links:
#             tar_socket = re_link.to_socket
#             links.new(out_socket, tar_socket)
#             delete_reroute(re_link, nodes, links, out_socket)
#         nodes.remove(reroute)

def split_all_and_merge_move(context, is_pre_merge=False):
    selected_nodes = context.selected_nodes
    if is_pre_merge:
        merge_group_input_linked(selected_nodes, context.active_node)
    selected_nodes = context.selected_nodes
    tree = context.space_data.edit_tree
    nodes = tree.nodes
    links = tree.links
    i = 0
    to_node_with_inputs = {}    # to_node_with_group_inputs节点输入接口的组输入数量
    for node in selected_nodes:
        if node.bl_idname == "NodeGroupInput" and node.select and has_link(node):
            i = -1
            for out_soc in node.outputs:
                if not out_soc.hide and out_soc.enabled and out_soc.bl_idname != "NodeSocketVirtual":
                    # 删掉组输入输出接口后面连的所有转接点
                    if out_soc.links:
                        for link in out_soc.links:
                            delete_reroute(link, nodes, links, out_soc)
                        
                    link_count = len(out_soc.links)
                    # soc_links = out_soc.links if out_soc.links else range(1)
                    if out_soc.links:
                        soc_links = out_soc.links 
                    else:
                        continue
                    for link in soc_links:
                        if hasattr(link, "is_valid") and not link.is_valid:       # 有些线存在，但是因为节点不同选项，线不可见
                            continue
                        i += 1
                        group_node = nodes.new('NodeGroupInput')
                        # group_node.location = node.location; group_node.location.y = node.location.y - 80 * i
                        group_id = {socket.identifier: socket for socket in group_node.outputs}
                        spec_in_socket = group_id[out_soc.identifier]
                        if link_count:         # is_linked:
                            links.new(spec_in_socket, link.to_socket)
                            to_node = spec_in_socket.links[0].to_node
                            to_socket = spec_in_socket.links[0].to_socket
                            if to_node in to_node_with_inputs:
                                to_node_with_inputs[to_node].append(group_node)
                            else:
                                to_node_with_inputs[to_node] = [group_node]
                            group_node.location = get_in_socket_location(to_node, to_socket)
                        for new_out_soc in group_node.outputs:
                            if new_out_soc.identifier != out_soc.identifier:
                                new_out_soc.hide = True
            nodes.remove(node)

    for node_list in to_node_with_inputs.values():
        if len(node_list) > 1:
            # 不知道为什么合并后会偏移md
            merge_group_input_linked(node_list, None, loc_at_max_y=True).location.x = abs_loc(node_list[0]).x

class W_OT_Split_All_And_Merge_Move(Operator):
    bl_idname = "cloud.split_all_and_merge_move"
    bl_label = trans("拆分组输入并移动合并")
    bl_description = trans("完全拆分组输入接口,并移动到连向的接口(to_socket),并合并连到一个节点上的和距离近的组输入节点")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        split_all_and_merge_move(context)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Merge_Node_And_Split_Merge_Move(Operator):
    bl_idname = "cloud.merge_node_and_split_merge_move"
    bl_label = trans("合并节点并拆分移动合并")
    bl_description = trans("先合并选中组输入节点，再完全拆分组输入接口,并移动节点们到连向的接口(to_socket),并合并 连到一个节点上的和距离近的组输入节点")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        split_all_and_merge_move(context, is_pre_merge=True)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_OT_Split_Group_Input_Linked(Operator):
    bl_idname = "cloud.split_group_input_linked"
    bl_label = trans("拆分组输入")
    bl_description = trans("拆分组输入接口,只留下连线的接口")
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return not False

    def execute(self, context):
        tree = context.space_data.edit_tree
        nodes = tree.nodes
        links = tree.links
        selected_nodes = context.selected_nodes
        # for node in tree.nodes:   # nodes 会包括后面新添加的节点
        #     for out_socket in node.outputs:
        #         if out_socket.is_linked:
        #             group_node = nodes.new('NodeGroupInput')
        #             for link in out_socket.links:
        #                 links.new(link.from_socket, link.to_socket)  # 新建的节点连上线，后面被判定is_linked继续拆分崩溃
        i = 0
        for node in selected_nodes:
            if node.bl_idname == "NodeGroupInput" and node.select:
                i = -1
                for out_socket in node.outputs:
                    if out_socket.is_linked:
                        i += 1
                        group_node = nodes.new('NodeGroupInput')
                        group_node.location = abs_loc(node)
                        group_node.location.y = abs_loc(node).y - 80 * i
                        group_id = {out_socket1.identifier: out_socket1 for out_socket1 in group_node.outputs}
                        for link in out_socket.links:
                            links.new(group_id[out_socket.identifier], link.to_socket)
                        for out_soc in group_node.outputs:
                            out_soc.hide = True
                nodes.remove(node)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)

class W_MT_Add_Group_Input_Hided_Socket(Menu):
    bl_idname = "W_MT_Add_Group_Input_Hided_Socket"
    bl_label = trans("添加组输入")

    @classmethod
    def poll(cls, context):
        return not (False)

    def draw(self, context):
        layout = self.layout
        get_socket_icon(layout)

class W_PT_Add_Group_Input_Hided_Socket(Panel):
    # bl_category = 'Group'
    bl_category = '节点树'
    bl_label = trans('w-组输入拆分')
    bl_idname = 'W_PT_Add_Group_Input_Hided_Socket'
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_context = ''
    bl_order = 4
    bl_options = {'DEFAULT_CLOSED'}
    bl_ui_units_x=0

    @classmethod
    def poll(cls, context):
        return not (False)

    def draw(self, context):
        layout = self.layout
        get_socket_icon(layout)

class W_PT_Add_New_Group_Item(Panel):
    bl_category = '节点树'
    bl_label = trans('w-添加组输入输出接口')
    bl_idname = 'W_PT_Add_New_Group_Item'
    bl_space_type = 'NODE_EDITOR'
    bl_region_type = 'UI'
    bl_context = ''
    bl_order = 4
    bl_options = {'DEFAULT_CLOSED'}
    bl_ui_units_x=0

    @classmethod
    def poll(cls, context):
        return not (False)

    def draw(self, context):
        # name = trans(socket_name)     # 什么时候加的?会出问题
        info = trans("添加输入输出接口")
        a_node = context.active_node
        if a_node and a_node.select and a_node.type=="GROUP":
            info = trans("给活动节点组添加接口")
        layout = self.layout
        layout.label(text=trans(info), icon='NODETREE')
        split = layout.split(factor=0.5, align=True)
        split.prop(context.scene, 'add_input_socket',  text=trans('输入接口'), toggle=True, icon='BACK')
        split.prop(context.scene, 'add_output_socket', text=trans('输出接口'), toggle=True, icon='FORWARD')
        get_icon_add_new_socket(layout, context)

class W_MT_Merge_Split_Move_Group_Input(Menu):
    bl_idname = "W_MT_Merge_Split_Move_Group_Input"
    bl_label = trans("组输入拆分合并移动")

    @classmethod
    def poll(cls, context):
        return not (False)

    def draw(self, context):
        layout = self.layout
        prefs = bpy.context.preferences.addons[__name__].preferences
        if bpy.app.version < (4, 0, 0):
            layout.operator('cloud.merge_node_and_split_merge_move', text=trans('合并节点并拆分'), icon="ANIM")
            layout.separator()
            layout.operator('cloud.split_all_and_merge_move', text=trans('拆分并移动合并'), icon="ANIM")
            layout.operator('cloud.split_all_and_move', text=trans('完全拆分并移动'), icon="ANIM")
            layout.operator('cloud.split_all_group_input_socket', text=trans('完全拆分组输入'), icon="")
            layout.separator()
            layout.operator('cloud.merge_group_input_linked',  text=trans('合并组输入连线'), icon="")
            layout.operator('cloud.split_group_input_linked', text=trans('拆分组输入连线'), icon="")
            layout.separator()
            layout.operator('cloud.merge_group_input_socket', text=trans('合并组输入接口'), icon="")
            layout.operator('cloud.split_group_input_socket', text=trans('拆分组输入接口'), icon="")
        else:
            layout.operator('cloud.merge_node_and_split_merge_move', text=trans('合并节点并拆分'), icon="ANIM")
            layout.separator()
            layout.operator('cloud.split_all_and_merge_move', text=trans('拆分并移动合并'), icon="ANIM")
            layout.operator('cloud.split_all_and_move', text=trans('完全拆分并移动'), icon="ANIM")
            if not prefs.simplify_menu:
                layout.operator('cloud.split_all_group_input_socket', text=trans('完全拆分组输入'), icon="SPLIT_HORIZONTAL")
            layout.separator()
            layout.operator('cloud.merge_group_input_linked',  text=trans('合并组输入连线'), icon="AREA_JOIN")
            layout.operator('cloud.split_group_input_linked', text=trans('拆分组输入连线'), icon="SPLIT_HORIZONTAL")
            layout.separator()
            if not prefs.simplify_menu:
                layout.operator('cloud.merge_group_input_socket', text=trans('合并组输入接口'), icon="AREA_JOIN")
                layout.operator('cloud.split_group_input_socket', text=trans('拆分组输入接口'), icon="SPLIT_HORIZONTAL")
                layout.separator()
            layout.operator('cloud.hide_group_input_sockets', text=trans('隐藏组输入接口'), icon="DECORATE")

classes = [
            W_OT_Add_Group_Input_Hided_Socket,
            W_MT_Add_Group_Input_Hided_Socket,
            W_PT_Add_Group_Input_Hided_Socket,
            W_OT_Add_New_Group_Item,
            W_PT_Add_New_Group_Item,
            W_OT_Merge_Group_Input_Socket,
            W_OT_Split_Group_Input_Socket,
            W_OT_Split_All_Group_Input_Socket,
            W_OT_Split_All_And_Move,
            W_OT_Merge_Group_Input_Linked,
            W_OT_Split_Group_Input_Linked,
            W_OT_Split_All_And_Merge_Move,
            W_OT_Hide_Group_Input_Sockets,
            W_OT_Merge_Node_And_Split_Merge_Move,
            W_MT_Merge_Split_Move_Group_Input,
            W_AddonPreferences,
            ]

def register():
    global _icons
    _icons = bpy.utils.previews.new()
    for png in png_list:
        _icons.load(png, os.path.join(os.path.dirname(__file__), 'icons', png), "IMAGE")

    for i in classes:
        bpy.utils.register_class(i)
    bpy.types.NODE_MT_editor_menus.append(w_add_to_node_mt_editor_menus_45569)
    bpy.types.Scene.add_input_socket  = bpy.props.BoolProperty(name='add_input_socket',  description='同时添加输入接口', default=True)
    bpy.types.Scene.add_output_socket = bpy.props.BoolProperty(name='add_output_socket', description='同时添加输出接口', default=False)

    kc = bpy.context.window_manager.keyconfigs.addon
    km = kc.keymaps.new(name='Node Editor', space_type='NODE_EDITOR')

    kmi = km.keymap_items.new('wm.call_menu', 'ONE', 'PRESS', ctrl=False, alt=False, shift=False, repeat=False)
    kmi.properties.name = 'W_MT_Add_Group_Input_Hided_Socket'
    addon_keymaps['A6EHS'] = (km, kmi)

    kmi = km.keymap_items.new('wm.call_panel', 'ONE', 'PRESS', ctrl=True, alt=True, shift=True, repeat=False)
    kmi.properties.name = 'W_PT_Add_Group_Input_Hided_Socket'
    kmi.properties.keep_open = True
    addon_keymaps['key_PT_Add_Group_Input'] = (km, kmi)

    kmi = km.keymap_items.new('wm.call_panel', 'ONE', 'PRESS', ctrl=True, alt=False, shift=False, repeat=False)
    kmi.properties.name = 'W_PT_Add_New_Group_Item'
    kmi.properties.keep_open = True
    addon_keymaps['key_Add_New_Group_Item'] = (km, kmi)

    kmi = km.keymap_items.new('wm.call_menu', 'ONE', 'PRESS', ctrl=False, alt=False, shift=True, repeat=False)
    kmi.properties.name = 'W_MT_Merge_Split_Move_Group_Input'
    addon_keymaps['key_Merge_Split_Move_Group_Input'] = (km, kmi)


def unregister():
    global _icons
    bpy.utils.previews.remove(_icons)
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    for km, kmi in addon_keymaps.values():
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    bpy.types.NODE_MT_editor_menus.remove(w_add_to_node_mt_editor_menus_45569)
    for i in classes:
        bpy.utils.unregister_class(i)